// -*- c++ -*-

#ifndef ATOMIC_H
#define ATOMIC_H

class simple_lock_t
{
public:
  bool test_and_set();
  void clear();
  void set();
  bool test();
private:
  int my_lock;
};

//
// inline stuff
//

// atomic operations

inline bool
compare_and_swap(unsigned *ptr, unsigned oldval,
		 unsigned newval)
{
  char ret;

  asm volatile
    ("cmpxchgl %1,(%2) ; sete %%al"
     : "=a" (ret)
     : "q" (newval), "q" (ptr), "0" (oldval)
     : "memory");

  return ret;
}

inline bool 
test_and_set(unsigned *l)
{
  int tmp;

  asm volatile
    ("xchg %0,%1" : "=r" (tmp) : "m" (*l), "0" (1)
     : "memory");

  return tmp;
}

// typesafe variants
template <class I>
inline bool compare_and_swap(I *ptr, I oldval, I newval)
{
  return compare_and_swap(reinterpret_cast<unsigned*>(ptr),
			  *reinterpret_cast<unsigned*>(&oldval),
			  *reinterpret_cast<unsigned*>(&newval));
}

template <class I>
inline bool test_and_set(I *l)
{
  return test_and_set(reinterpret_cast<unsigned*>(l));
}

// simple_lock_t

#if 0
inline bool simple_lock_t::test_and_set()
{
  int tmp;

  asm volatile
    ("xchg %0,%1" : "=r" (tmp) : "m" (my_lock), "0" (1));

  return tmp;
}
#else
inline bool simple_lock_t::test_and_set()
{
  return ::test_and_set(&my_lock);
}
#endif


inline void simple_lock_t::clear()
{
  my_lock = 0;
}

inline void simple_lock_t::set()
{
  my_lock = 1;
}

inline bool simple_lock_t::test()
{
  return my_lock;
}

// XXX

#endif
