// -*- c++ -*-

#ifndef CONFIG_H
#define CONFIG_H

#define THREAD_BLOCK_SIZE (0x800)

#ifndef ASSEMBLER

#include <termios.h>
#include <flux/tty.h>

#include <l4/sys/types.h>

// special magic to allow old compilers to inline constants
#if __GNUC__ <= 2 && __GNUC_MINOR__ <= 90
# define MAGIC_CONST(x)
# define CONST_TRUE 1
# define CONST_FALSE 0
#else
# define MAGIC_CONST(x) x
# define CONST_TRUE true
# define CONST_FALSE false
#endif

class config
{
public:
  // initialization
  static void init();

  // global kernel configuration
  static const unsigned kernel_version_id = 0x01004444; // "DD\000\001"

  static const char *const kernel_version_string = 
    "Welcome to Fiasco!\n"
    "DD-L4/x86 microkernel (c) 1998 TU Dresden - " __DATE__;
  // don't change the part "DD-L4/x86 microkernel" -- Rmgr depends on it.

  static const unsigned kernel_mem_per_cent = 20;

  static const unsigned thread_block_size = THREAD_BLOCK_SIZE;

  static const unsigned thread_block_mask 
    = (thread_block_size - 1) | ~((thread_block_size << (11+7)) - 1);
  static const unsigned thread_block_id_factor = thread_block_size/0x400;
  
  static const MAGIC_CONST(bool) backward_compatibility = CONST_TRUE;
  static const MAGIC_CONST(bool) hlt_works_ok = CONST_TRUE;

  static const MAGIC_CONST(bool) conservative = CONST_FALSE;

  static bool use_hercules, use_serial;

  static termios* const serial_termios = &base_raw_termios;
  static unsigned serial_com_port;

#ifdef PROFILE
  static bool profiling;
#else
  static const MAGIC_CONST(bool) profiling = CONST_FALSE;
#endif
  static const int profiling_rate = 100;
  static const int profile_irq = 0;

  // kernel (idle) task definitions
  static const unsigned kernel_taskno = 0;
  static const l4_threadid_t kernel_id;
  static const unsigned kernel_prio = 1;
  static const unsigned kernel_mcp = 255;
  
  // sigma0 task definitions
  static const unsigned sigma0_taskno = 2;
  static const l4_threadid_t sigma0_id; 
  static const unsigned sigma0_prio = 0x10;
  static const unsigned sigma0_mcp = 0;

  // root (boot) task definitions
  static const unsigned boot_taskno = 4;
  static const l4_threadid_t boot_id;
  static const unsigned boot_prio = 0x10;
  static const unsigned boot_mcp = 255;

  // default number of clock ticks / time slice
  static const int default_time_slice = 10;
  // one clock tick takes this many microsecs
  static const unsigned microsec_per_tick = 976LL;
};

#endif /* ! ASSEMBLER */

#endif
