// -*- c++ -*-

#ifndef JDB_H
#define JDB_H

#include <stdio.h>
#include <flux/gdb.h>

#include "thread.h"
#include "kmem.h"
#include "config.h"

struct termios;


class jdb
{
public:
  static void init();
  static int enter_kdebug(trap_state *state);
  static int source_level_debugging() { return use_nested; };
  static bool connected();

private:
  enum command_t { thread_c, breakpoint_c, display_c };
  typedef   struct {
    enum command_t command;
    l4_threadid_t id;
    unsigned long address;
    int size;
  } cmd_t;
  typedef struct {
    unsigned long address;
    char len, mode;
  } breakpoint_t;
  typedef struct {
    char name[3];
    char error_code;
  } traps_t;
  
  static bool _connected;

  static unsigned int pic_mask;
  static char show_state, use_nested;
  static thread_t *current_present, *current_ready;
  static breakpoint_t bps[4];
  static char *mode_names[4];
  static traps_t traps[19];
  static unsigned debug_control;

  static char* debug_msg;
  static int msg_len;

  jdb();			// default constructors are undefined
  jdb(const jdb&);

  static int (*nested_trap_handler)(trap_state *state);
  static cmd_t cmd;

  static void open_debug_console(trap_state *state);
  static void close_debug_console(void);

  static int execute_command(trap_state *state);


  /* debug register manipulation */
  static void set_debug_address_register(unsigned long address, int reg_num);

  static void reset_debug_status_register() 
  { __asm__("movl %0, %%db6" : /* no output */ : "r" (0));};
  static unsigned long get_debug_status_register() 
  { unsigned long st; __asm__("movl %%db6, %0" : "=r" (st)); return st;};

  static unsigned long get_debug_control_register()
  { unsigned long ctrl; __asm__("movl %%db7, %0" : "=r" (ctrl)); return ctrl;};
  static void set_debug_control_register(unsigned long ctrl)
   { __asm__("movl %0, %%db7" : /* no output */ : "r" (ctrl)); };

  static void set_breakpoint(int mode);
  static void show_help(void);

  /* basic output functions */
  static void put_character(int c);
  static void cursor(int y, int x);
  static void clear_screen();
  static inline void home() { putchar('\001'); };
  static void clear_to_eol(){ putchar('\005'); };;

  /* basic input functions */
  static int get_character();
  static l4_threadid_t get_thread_id(void);
  static int get_32();
  static int get_num(int mask);

  /* breakpoint and single step */
  static void single_step(trap_state *state);
  static void break_point(trap_state *state);
  static void list_breakpoints();
  static void delete_breakpoint();

  /* complex io functions */
  static int switch_debug_state(trap_state *state);
  static void show_tcb(l4_threadid_t, trap_state *state);
  static void show_present_or_ptab(trap_state *state);
  static void show_ready(trap_state *state);
  static void show_kernel_state();
  static void page_fault_protocol();

  friend class _foo;		// avoids warning "no public cons/no friends"
};

#define jdb_enter_kdebug(text) \
asm(\
    "int	$3	\n\t"\
    "jmp	1f	\n\t"\
    ".ascii	\""text "\"\n\t"\
    "1:			\n\t"\
    "nop;nop;nop			\n\t"\
    )

#endif
