// -*- c++ -*-

#ifndef SLAB_H
#define SLAB_H

class slab;

// The anonymous slab allocator.  You can specialize this allocator by
// providing your own initialization functions and your own low-level
// allocation functions.
class slab_cache_anon
{
public:
  slab_cache_anon(unsigned long slab_size, unsigned elem_size, 
		  unsigned alignment);
  virtual ~slab_cache_anon();

  virtual void *alloc();	// request initialized member from cache
  virtual void free(void *cache_entry);	// return initialized member to cache

  virtual bool reap();		// request that cache returns memory to system

protected:
  friend class slab;

  // Low-level allocator functions:

  // Allocate/free a block.  "size" is always a multiple of PAGE_SIZE.
  virtual void *block_alloc(unsigned long size, unsigned long alignment) = 0;
  virtual void block_free(void *block, unsigned long size) = 0;

  // Element initialization functions:
  virtual void elem_ctor(void *elem);
  virtual void elem_dtor(void *elem);

  void destroy();		// descendant should call this in destructor
  
private:
  slab_cache_anon();
  slab_cache_anon(const slab_cache_anon&); // default constructor is undefined

  //
  // data declaration follows
  // 

  // The slabs of this cache are held in a partially-sorted
  // doubly-linked list.  First come the fully-active slabs (all
  // elements in use), then the partially active slabs, then empty
  // slabs.
  slab *_first_slab, *_first_available_slab, *_last_slab;
  unsigned long _slab_size;
  unsigned _elem_size, _latest_offset, _alignment;
};				// end declaration of class slab_cache

#endif // SLAB_H
