// -*- c++ -*-

#ifndef LOCK_H
#define LOCK_H

#include "thread.h"
#include "globals.h"
#include "atomic.h"

// 
// Helping lock.  This is a lock implementation based on Fiasco's
// transaction mechanism.
// 
class helping_lock_t
{
public:
  helping_lock_t();

  void lock();
  bool test_and_set();
  void clear();
  bool test() const;
  thread_t *owner() const;

private:
  void lock_current();
  void unlock_current();

  thread_t *_owner;
  bool _in_trans;
};

// 
// a guard object using a lock such as helping_lock_t
// 
template<class LOCK>
class lock_guard_t
{
public:
  lock_guard_t(LOCK *l);
  ~lock_guard_t();

private:
  LOCK *_lock;
};

typedef lock_guard_t<helping_lock_t> helping_lock_guard_t;

// 
// helping_lock_t inlines 
// 
inline
helping_lock_t::helping_lock_t()
  : _owner(0)
{
}

inline void helping_lock_t::lock()
{
  do 
    {
      thread_t *o = _owner;
      if (o)
	o->trans_help();
    }
  while (test_and_set());
}

inline bool helping_lock_t::test() const
{
  return _owner;
}

inline thread_t *helping_lock_t::owner() const
{
  return _owner;
}

// 
// lock_guard_t inlines
// 
template<class LOCK>
inline
lock_guard_t<LOCK>::lock_guard_t(LOCK *l)
  : _lock(l)
{
  _lock->lock();
}

template<class LOCK>
inline
lock_guard_t<LOCK>::~lock_guard_t()
{
  _lock->clear();
}

#endif // LOCK_H
