/*
 * Functions for communication with sigma 0
 */

#include <stdio.h>
#include <oskit/page.h>
#include <oskit/x86/paging.h>

#include <l4/sys/types.h>
#include <l4/sys/ipc.h>
#include <l4/sys/syscalls.h>

#include "sigma0.h"

/* #define DEBUG */

static l4_threadid_t sigma0_id = L4_INVALID_ID;
static int sigma0_initialized = 0;

#if 0
l4_kernel_info_t *l4_kernel_info = 0;
#endif

static void
check_pager(void)
{
  /* get pager id if required */
  if (!sigma0_initialized)
    {
      l4_threadid_t pager = L4_INVALID_ID, preempter = L4_INVALID_ID;
      dword_t eip = -1, esp = -1, eflags;
      l4_thread_ex_regs(l4_myself(), eip, esp, 
			&preempter, &pager, &eflags, &eip, &esp);
      if (!l4_is_invalid_id(pager))
	{
	  sigma0_id = pager;
	  sigma0_initialized = 1;
	}
      else
	enter_kdebug("check_pager_id: invalid pager");
    }
  
}

static int 
call_pager(dword_t dw0, dword_t dw1, 
	   l4_msgdope_t *map_msg, dword_t *reply_dw0, dword_t *reply_dw1, 
	   l4_timeout_t timeout, l4_msgdope_t *result)
{
  int err;

  check_pager();
  do 
    {
      err = l4_i386_ipc_call(sigma0_id, 
			     L4_IPC_SHORT_MSG, dw0, dw1,
			     map_msg, reply_dw0, reply_dw1,
			     timeout, result);
    }
  while ((err & ~L4_IPC_SND_ERR_MASK) == L4_IPC_RECANCELED);

  return err;
}

#if 0
int 
setup_kernel_info_page(void)
{
  l4_msgdope_t result;
  l4_snd_fpage_t fpage;
  int error;

  l4_kernel_info = (l4_kernel_info_t *)0;
  /* request kernel info page page from sigma 0 */
  error = call_pager(1, 1,
		     L4_IPC_MAPMSG(KERNEL_INFO_PAGE_ADDRESS, L4_LOG2_PAGESIZE), 
		     &fpage.snd_base, &fpage.fpage.fpage,  
		     L4_IPC_NEVER, &result);
  if (!error)
    {
      if (l4_ipc_fpage_received(result))
	{
	  l4_kernel_info = (l4_kernel_info_t *)KERNEL_INFO_PAGE_ADDRESS;
	  return 0;
	}
      else
	panic("setup_kernel_info_page:" 
	       " sigma 0 denied mapping of kernel info page\n");
    }
  else
    panic("setup_kernel_info_page:"
	   " ipc error %x while requesting kernel info page\n", error);
  return 1; 
}
#endif

static int 
check_error(int error, l4_msgdope_t result, l4_snd_fpage_t fpage)
{
  if (!error)
    {
      if (l4_ipc_fpage_received(result))
	{
	  if (l4_ipc_is_fpage_writable(fpage.fpage))
	    {
	      if (fpage.fpage.fp.size == RP_4MB)
		return RP_4MB; 
	      else
		return RP_WRITABLE;
	    }
	  else
	    {
	      return RP_NOT_WRITABLE;
	    }
	}
      else
	{
	  return RP_NOT_MAPPED;
	}
    }
  else
    {
      printf("sigma0_request_page: ipc error: 0x%x\n", error);
    }
  return RP_NOT_MAPPED;
} 

int 
sigma0_request_page(oskit_addr_t address)
{
  l4_msgdope_t result;
  l4_snd_fpage_t fpage;
  int error;

  error = call_pager(address+1, L4_LOG2_SUPERPAGESIZE*4,
		     L4_IPC_MAPMSG(address, L4_LOG2_SUPERPAGESIZE), 
		     &fpage.snd_base, &fpage.fpage.fpage,  
		     L4_IPC_NEVER, &result);
  return check_error(error, result, fpage);
}

int 
sigma0_request_dev_page(oskit_addr_t virt_address, 
			oskit_addr_t phys_address)
{
  l4_msgdope_t result;
  l4_snd_fpage_t fpage;
  int error;

  virt_address &= ~SUPERPAGE_MASK;
  phys_address &= ~SUPERPAGE_MASK;

  /* request area from sigma 0 */
  error = call_pager(phys_address - 0x40000000, 0,
		     L4_IPC_MAPMSG(virt_address, L4_LOG2_SUPERPAGESIZE), 
		     &fpage.snd_base, &fpage.fpage.fpage,  
		     L4_IPC_NEVER, &result);
  return check_error(error, result, fpage);
}
