#include <oskit/dev/dev.h>
#include <oskit/dev/osenv_sleep.h>

#include <l4/sys/types.h>
#include <l4/sys/ipc.h>
#include <l4/sys/syscalls.h>

/* hack: symbols that can be referenced to suck in this object */
asm("l4_oskit_support_osenv_sleep:");

enum { State_going_to_sleep = 0, State_woken_up, State_max};

/* helpers */

L4_INLINE int compare_and_swap(unsigned *ptr, unsigned oldval,
                               unsigned newval);

L4_INLINE int
compare_and_swap(unsigned *ptr, unsigned oldval,
                 unsigned newval)
{
  char ret;

  asm volatile
    ("cmpxchgl %1,(%2) ; sete %%al"
     : "=a" (ret)
     : "q" (newval), "q" (ptr), "0" (oldval)
     : "memory");

  return ret;
}

/* The osenv_sleep interface */

void
osenv_sleep_init(osenv_sleeprec_t *sr) 
{
  osenv_assert(sr);
  
  sr->data[0] = OSENV_SLEEP_WAKEUP;   /* Return code for osenv_sleep() */
  sr->data[1] = (void *) State_going_to_sleep;
}

int
osenv_sleep(osenv_sleeprec_t *sr) 
{
  /* We won't get anywhere if interrupts aren't enabled! */
  osenv_intr_enable();

  /* The following code tests whether we should go to sleep, and if
     so, just sleeps forever -- but before doing so, it helps
     osenv_wakeup() by putting the address of our return EIP into
     sr->data[1].  osenv_wakeup() will later poke the address of
     sleep_exit to that address to avoid a race. */
  /* XXX: Assumption about wrap_deliver_signal()'s stack frame (irq.c) */
  asm volatile
    ("pushl %%ebp
      pushl %%ecx
      leal  -4(%%esp), %%ebx
      cmpxchgl %%ebx, (%%ecx)
      jne   sleep_exit
1:    xorl  %%ecx,%%ecx
      xorl  %%ebp,%%ebp
      xorl  %%esi,%%esi
      xorl  %%edi,%%edi
      leal  -1(%%ecx), %%eax
      int   $0x30
      movl  (%%esp), %%ecx
      leal  -4(%%esp), %%ebx
      cmpl  %%ebx,(%%ecx)
      je    1b
sleep_exit: 
      addl  $4,%%esp
      popl  %%ebp"
     : 
     : "a" (State_going_to_sleep), "c" (sr->data + 1)
     : "eax", "ebx", "ecx", "edx", "esi", "edi");
      
  return (int) (sr->data[0]);
}

void sleep_exit(void);		/* contained in osenv_sleep() above */

void
osenv_wakeup(osenv_sleeprec_t *sr, int wakeup_status) 
{
  for(;;)
    {
      void *sleep_program_counter;

      osenv_assert((unsigned)(sr->data[1]) != State_woken_up);

      if (compare_and_swap((unsigned*) (sr->data + 1), 
			   State_going_to_sleep, State_woken_up))
	break;

      sleep_program_counter = sr->data[1];

      if ((unsigned) sleep_program_counter >= State_max
	  && compare_and_swap((unsigned*) (sr->data + 1),
			      (unsigned) sleep_program_counter, 
			      State_woken_up))
	{
	  /* make the sleep return, even if it hasn't started its IPC
             yet */
	  *(void **)sleep_program_counter = sleep_exit;
	  break;
	}
    }

  sr->data[0] = (void*) wakeup_status;
}
